/*
 *	Window functions. Slow slow slow slow mem hungry & slow...
 */

#include <stdlib.h>
#include <stdio.h>
#include <vga.h>
#include <vgagl.h>
#include <vgamouse.h>
#include <string.h>

#include "hr_defines.h"
#include "hr_colors.h"
#include "hr_zones.h"
#include "hr_bobs.h"
#include "hr_mouse.h"
#include "hr_buttons.h"
#include "hr_general.h"
#include "hr_bitmaps.h"
#include "hr_input_handler.h"
#include "hr_fonts.h"
#include "hr_graphics.h"

struct hr_window_gadgets{
	int button_n;
	struct hr_button **button;
};


struct hr_window {
	struct hr_window *next;
	int flags;
	struct hr_color *colors;
	struct hr_bitmap *background;
	struct hr_bitmap *foreground;
	struct hr_zone *zone;
	struct hr_window_gadgets *gadgets;
};



/* Returns pointer to frontmost (last) window in window list.
 * If there are no windows open, NULL is returned.*/
struct hr_window *hr_frontmost_window()
{
	struct hr_window *wp;
	wp=hr.firstwindow;
	if(wp!=NULL){
		while(wp->next!=NULL)
		  wp=wp->next;
	}
	return wp;
}


/* Removes all windows from-frontmost-to-wp from screen.
 * This must be called before moving,
 * resizing or in any way modifying windows. It is ok to modify CONTENTS
 * of frontmost window tho.*/
void hr_get_windows(struct hr_window *wp)
{
	struct hr_window *t;
	int gotcha,width,height;
	int cx1,cy1,cx2,cy2;
	/* In reverse order, "get" windows.*/
	if(wp!=NULL){
		cx1=__clipx1;
		cy1=__clipy1;
		cx2=__clipx2;
		cy2=__clipy2;
		gl_setclippingwindow(0,0,WIDTH-1,HEIGHT-1);
		while(!(wp->flags & WINDOW_STORED)){
			t=wp;
			width=(t->zone->x2)-(t->zone->x1)+1;
			height=(t->zone->y2)-(t->zone->y1)+1;
			gotcha=0;
			while((t->next!=NULL) && (!gotcha)){
				if(t->next->flags & WINDOW_STORED)
				  gotcha=1;
				t=t->next;
			}
			t->flags |= WINDOW_STORED;
			/* Save the window contents. */
			hr_delete_bitmap(t->foreground);
			t->foreground=hr_create_bitmap(malloc(width*height),
													 NULL,width,height,
													 0,GFX_COLOR);
			gl_getbox(t->zone->x1,t->zone->y1,
						 width,height,(t->foreground->data));
			/* Restore gfx under window, free mem.*/
			if(t->background!=NULL){
				gl_putbox(t->zone->x1,t->zone->y1,
							 width,height,
							 (t->background->data));
				hr_delete_bitmap(t->background);
				t->background=NULL;
			}
		}
		gl_setclippingwindow(cx1,cy1,cx2,cy2);
	}	
}
	

/* Puts all windows back to their places, frees foregrounds. Call this
 * After moving, resizing or otherwise modifying any non-frontmost
 * window.*/
void hr_put_windows()
{
	struct hr_window *wp;
	int width,height;
	int cx1,cy1,cx2,cy2;
	/**/
	cx1=__clipx1;
	cy1=__clipy1;
	cx2=__clipx2;
	cy2=__clipy2;
	gl_setclippingwindow(0,0,WIDTH-1,HEIGHT-1);
	/* In reverse order, "get" windows.*/
	wp=hr.firstwindow;
	while(wp!=NULL){
		if(wp->flags & WINDOW_STORED){
			wp->flags &= ~(WINDOW_STORED);
			width=(wp->zone->x2)-(wp->zone->x1)+1;
			height=(wp->zone->y2)-(wp->zone->y1)+1;
			/* Save the background */
			hr_delete_bitmap(wp->background);
			wp->background=hr_create_bitmap(malloc(width*height),
													  NULL,width,height,
													  0,GFX_COLOR);
			gl_getbox(wp->zone->x1,wp->zone->y1,
						 width,height,
						 wp->background->data);
			/* Draw window contents and free foreground bitmap.
			 * It is ok for a window to be resized.*/
			if(wp->foreground!=NULL){
				if((width==(wp->foreground->width))&&
					(height==(wp->foreground->height)))
				  gl_putbox(wp->zone->x1,wp->zone->y1,
								width,height,wp->foreground->data);
				else
				  gl_putboxpart(wp->zone->x1,wp->zone->y1,
									 width,height,
									 (wp->foreground->width),
									 (wp->foreground->height),
									 (wp->foreground->data),0,0);
				hr_delete_bitmap(wp->foreground);
				wp->foreground=NULL;
			}
		}
		/* Neeeeeeext! */
		wp=wp->next;
	}
	gl_setclippingwindow(cx1,cy1,cx2,cy2);
}


/* Redraws bevelbox, titlebac and stuff. Only works for frontmost window.*/
void hr_window_refresh(struct hr_window *wp)
{
	int bflags,l;
	int x1=0,x2=0,y1=0,y2=0,width=0,height=0,width2;
	struct hr_button *b;
	if(wp!=NULL){
		width=(wp->zone->x2)-(wp->zone->x1)+1;
		height=(wp->zone->y2)-(wp->zone->y1)+1;
		/* Remove windows on top of this one.*/
		if(wp!=hr_frontmost_window())
		  hr_get_windows(wp->next);
		/* Clear window area. */
		gl_fillbox((wp->zone->x1)+1,(wp->zone->y1)+1,
					  width-2,height-2,
					  hr_get_color_index(wp->colors,COLOR_BACKGROUND));
		if(!(wp->flags & WINDOW_NOBORDER)){
			/* Draw edge bevelbox.*/
			hr_bevelbox(wp->zone->x1,wp->zone->y1,
							wp->zone->x2,wp->zone->y2,
							wp->colors,NONE);
		}
		if(wp->flags & WINDOW_INNERBOX){
			x1=(wp->zone->x1)+2;
			x2=(wp->zone->x2)-2;
			y1=(wp->zone->y1)+2;
			y2=(wp->zone->y2)-2;
		}
		/* Set the window button positions and turn the buttons visible.*/
		if((wp->gadgets)!=NULL){
			if(wp->gadgets->button_n>0){
				width2=((width-3)/(wp->gadgets->button_n));
				/* Now put all the buttons in a row.*/
				for(l=0;l<(wp->gadgets->button_n);l++){
					b=(wp->gadgets->button)[l];
					if(b!=NULL) {
						b->zone->x1=(wp->zone->x1)+2+l*width2;
						b->zone->x2=(wp->zone->x1)+2+(l+1)*width2-2;
						b->zone->y1=(wp->zone->y1)+height-7-(b->font->height);
						if(wp->flags & WINDOW_INNERBOX){
							if((b->zone->y1)<(y2+2))
							  y2=(b->zone->y1)-2;
						}
						b->zone->y2=(wp->zone->y2)-2;
						/* Make the button visible */
						bflags=(b->flags &= ~(BUTTON_INVISIBLE));
						hr_update_button(b,UPDATE_FLAGS,&bflags);
					}
				}
			}
		}
		if(wp->flags & WINDOW_INNERBOX){
			/* Draw inner bevelbox.*/
			hr_bevelbox(x1,y1,x2,y2,wp->colors,BUTTON_PRESSED);
		}
		/* Restore windows on top of this one */
		hr_put_windows();
	}
}


/* Opens (creates) a window */
struct hr_window *hr_window_open(int x,int y,int w,int h,
											struct hr_color *colors,
											int flags)
{
	struct hr_window *wp;
	struct hr_button *b;
	/* Redraw old default button.*/
	if((b=hr.default_button)){
		hr.default_button=NULL;
		hr_update_button(b,UPDATE_GFXONLY,NULL);
	}
	/* Create new window struct */
	hr.active_window=wp=malloc(sizeof(struct hr_window));
	wp->flags=flags;
	wp->zone=hr_create_zone(x,y,x+w-1,y+h-1,ZONE_WINDOW);
	wp->colors=colors;
	/* Nullify foreground pointer */
	wp->foreground=NULL;
	/* Save the background */
	wp->background=hr_create_bitmap(malloc(w*h),NULL,w,h,0,GFX_COLOR);
	gl_getbox(x,y,w,h,wp->background->data);	
	/* Link buttons to window.*/
	wp->gadgets=NULL;
	/* Set some pointers.*/
	if(hr.firstwindow==NULL)
	  hr.firstwindow=wp;
	else
	  hr_frontmost_window()->next=wp;
	wp->next=NULL;
	/* Clipping window */
	if(flags & WINDOW_CLIPPING)
	  gl_setclippingwindow(wp->zone->x1,wp->zone->y1,
								  wp->zone->x2,wp->zone->y2);
	/* Draw it.*/
	hr_window_refresh(wp);
	/* Awayyy......*/
	return wp;
}


/* Closes a window. Since possible window buttons were (hopefully!)
 * initialized as INVISIBLE, the background is redrawn correctly.*/
void hr_window_close(struct hr_window *wp)
{
	struct hr_window *t;
	struct hr_button *b;
	int l;
	if(wp!=NULL) {
		/* Remove windows on top of wp.*/
		hr_get_windows(wp);
		/* Delete window zone.*/
		hr_delete_zone(wp->zone);
		/* Delete the window buttons. It is ok to remove NULL buttons.*/
		if(wp->gadgets!=NULL){
			for(l=0;l<(wp->gadgets->button_n);l++)
			  hr_delete_button((wp->gadgets->button)[l]);
			free(wp->gadgets->button);
			free(wp->gadgets);
		}
		/* it is ok to free/delete NULLs...*/
		hr_delete_bitmap(wp->background);
		hr_delete_bitmap(wp->foreground);
		/* Next-pointers must "jump over" *wp.*/
		if(wp==hr.firstwindow)
		  hr.firstwindow=wp->next;
		else{
			t=hr.firstwindow;
			while((t->next)!=wp)
			  t=t->next;
			t->next=(wp->next);
		}
		free(wp);
		/* Redraw windows on top of *wp if necessary */
		hr_put_windows();
	}
	/* Set active window */
	hr.active_window=hr_frontmost_window();
	/* Clipping window */
	if(hr.active_window==NULL)
	  gl_setclippingwindow(0,0,WIDTH-1,HEIGHT-1);
	else
	  gl_setclippingwindow(hr.active_window->zone->x1,
								  hr.active_window->zone->y1,
								  hr.active_window->zone->x2,
								  hr.active_window->zone->y2);
	/* Set default button - ie. found button that has BUTTON_DEFAULT_BUTTON flag
	 * set and that belongs to active_window.*/
	b=hr.firstbutton;
	while(b!=NULL){
		if((b->flags & BUTTON_DEFAULT_BUTTON) &&
			((b->parent_window)==hr.active_window) &&
			(b->flags & BUTTON_ENABLED) &&
			!(b->flags & BUTTON_INVISIBLE) &&
			!(b->flags & BUTTON_NO_TAB_CYCLE))
		  {
			  hr.default_button=b;
			  hr_update_button(b,UPDATE_GFXONLY,NULL);
			  break;
		  }
		b=b->next;
	}
}


/* Moves a window */
void hr_window_move(struct hr_window *wp,int nx,int ny)
{
	if(wp!=NULL){
		hr_get_windows(wp);
		wp->zone->x1=nx;
		wp->zone->y1=ny;
		hr_put_windows();
	}
}


/* Resizes a window */
void hr_window_resize(struct hr_window *wp,int nwidth,int nheight)
{
	if(wp!=NULL){
		hr_get_windows(wp);
		wp->zone->x2=(wp->zone->x1)+nwidth-1;
		wp->zone->y2=(wp->zone->y1)+nheight-1;
		hr_put_windows();
	}
}


/* Moves a window to front.*/
void hr_window_to_front(struct hr_window *wp)
{
	struct hr_window *t;
	if((wp!=NULL) && (wp!=hr_frontmost_window())){
		hr_get_windows(wp);
		if(wp==hr.firstwindow)
		  hr.firstwindow=hr.firstwindow->next;
		else{
			t=hr.firstwindow;
			while(t->next!=wp)
			  t=t->next;
			t->next=(t->next->next);
		}
		hr_frontmost_window()->next=wp;
		wp->next=NULL;
		hr_put_windows();
	}
}


/* Moves a window to back */
void hr_window_to_back(struct hr_window *wp)
{
	struct hr_window *t;
	if((wp!=NULL)&&(wp!=hr.firstwindow)){
		hr_get_windows(wp);
		t=hr.firstwindow;
		while(t->next!=wp)
		  t=t->next;
		t->next=t->next->next;
		wp->next=hr.firstwindow;
		hr.firstwindow=wp;
		hr_put_windows();
	}
}


/* Creates one window gadget struct and zeroes it.*/
struct hr_window_gadgets *hr_create_window_gadgets(int button_n)
{
	struct hr_window_gadgets *ggg;
	ggg=malloc(sizeof(struct hr_window_gadgets));
	ggg->button_n=button_n;
	if(button_n>0)
	  ggg->button=malloc(button_n*sizeof(struct hr_button *));
	else
	  ggg->button=NULL;
	return ggg;
}


/* Returns pointer to window under (x,y) */
struct hr_window *hr_get_window(int x,int y)
{
	struct hr_window *win=hr.firstwindow,*xy_win=NULL;
	while(win!=NULL){
		if((x>=(win->zone->x1))&&
			(x<=(win->zone->x2))&&
			(y>=(win->zone->y1))&&
			(y<=(win->zone->y2)))
		  xy_win=win;
		win=win->next;
	}
	return xy_win;
}
